@@ -7,7 +7,7 @@ import random
 
 import requests
 from MammothAnalytics import const
-from MammothAnalytics.const import USER_PERMISSIONS
+from MammothAnalytics.const import USER_PERMISSIONS,RESERVED_BATCH_COLUMN_INTERNAL_NAMES_AND_KEYS
 from MammothAnalytics.errors import AuthError, UnknownError
 from MammothAnalytics.errors import MalformedRequestError, AuthenticationError, \
     NotFoundError, AuthorizationError
@@ -18,6 +18,9 @@ log = logging.getLogger(__name__)
 logging.basicConfig(level=logging.DEBUG)
 logging.getLogger("requests").setLevel(logging.WARNING)
 
+MAX_RETRIES = 50
+RETRY_DELAY_IN_SEC = 2
+
 
 def handleError(f):
     def new_function(*args, **kwargs):
@@ -406,7 +409,8 @@ class MammothConnector(object):
         return response.json()['id']
 
     @handleError
-    def append_datasets(self, source_dataset_id, target_dataset_id, column_mapping, change_map=None, new_ds_params=None):
+    def append_datasets(self, source_dataset_id, target_dataset_id, column_mapping, change_map=None,
+                        new_ds_params=None):
         """
         Method to combine two datasets
         """
@@ -430,7 +434,8 @@ class MammothConnector(object):
         :param file_id:
         :return: A dataset dictionary
         """
-        while True:
+        retry_count = 0
+        while retry_count < MAX_RETRIES*4:
             response = self._make_signed_request('get', '/files/{0}'.format(file_id))
             file_info = response.json()['file']
             status = file_info['status']
@@ -438,7 +443,10 @@ class MammothConnector(object):
                 log.info(file_info)
                 break
             else:
-                time.sleep(1)
+                time.sleep(RETRY_DELAY_IN_SEC)
+            retry_count += 1
+        if retry_count == MAX_RETRIES*4:
+            raise RuntimeError(f"Max time limit reached to get processed file .")
         response = self._make_signed_request('get', '/files/{0}'.format(file_id))
         file_info = response.json()['file']
         final_ds_id = file_info["additional_data"].get("final_ds_id")
@@ -457,19 +465,23 @@ class MammothConnector(object):
 
     @handleError
     def _wait_for_ds_status(self, ds_id, status, check_for_status_equality=True):
-        while ds_id and True:
+        retry_count = 0
+        while ds_id and retry_count < MAX_RETRIES*4:
             ds = self.get_dataset(ds_id)
             ds_status = ds['status']
             if check_for_status_equality:
                 if ds_status == status:
                     break
                 else:
-                    time.sleep(1)
+                    time.sleep(RETRY_DELAY_IN_SEC)
             else:
                 if ds_status != status:
                     break
                 else:
-                    time.sleep(1)
+                    time.sleep(RETRY_DELAY_IN_SEC)
+            retry_count += 1
+        if retry_count == MAX_RETRIES * 4:
+            raise RuntimeError(f"Max time limit reached in _wait_for_ds_status")
 
     @handleError
     def sync_now(self, ds_id):
@@ -579,12 +591,29 @@ class MammothConnector(object):
         return r.json()
 
     @handleError
+    def wait_for_view_to_finish_processing(self, view_id):
+        retry_count = 0
+        while retry_count < MAX_RETRIES * 2:
+            if retry_count != 0:
+                time.sleep(RETRY_DELAY_IN_SEC)
+            view = self.get_view_details(view_id)
+            if view['status'] != 'processing':
+                break
+            retry_count += 1
+        if retry_count == MAX_RETRIES * 2:
+            raise RuntimeError(f"Max retry limit reached to get view in ready state")
+    @handleError
     def wait_till_view_is_ready(self, view_id):
-        while True:
-            time.sleep(1)
+        retry_count = 0
+        while retry_count < MAX_RETRIES*2:
+            if retry_count != 0:
+                time.sleep(RETRY_DELAY_IN_SEC)
             view = self.get_view_details(view_id)
             if view['status'] == 'ready':
                 break
+            retry_count += 1
+        if retry_count == MAX_RETRIES*2:
+            raise RuntimeError(f"Max retry limit reached to get view in ready state")
 
     @handleError
     def copy_template_from_view(self, ds_id, view_id):
@@ -647,13 +676,17 @@ class MammothConnector(object):
         return response['request_id']
 
     def _wait_till_request_is_complete_get_data(self, view_id, request_id):
-        while True:
-            time.sleep(1)
+        retry_count = 0
+        while retry_count < MAX_RETRIES:
+            if retry_count != 0:
+                time.sleep(RETRY_DELAY_IN_SEC)
             r = self._make_signed_request('get','/workspaces/{0}/requests/{1}'.format(view_id, request_id))
             data = r.json()
             log.debug("data in _wait_till_request_is_complete_get_data {}".format(data))
             if data['status'] in ['success', 'failure', 'referror', 'error']:
                 return data
+            retry_count += 1
+        raise RuntimeError(f"Max retry limit reached to get workspace out of processing .")
 
     @handleError
     def list_actions(self, view_id):
@@ -740,16 +773,22 @@ class MammothConnector(object):
         url = '/integrations/{0}/identities'.format(integration_key)
         r = self._make_signed_request('post', url, json=identity_config)
         request_id = r.json()['request_id']
-        while True:
-            time.sleep(2)
+        retry_count = 0
+        while retry_count < MAX_RETRIES:
+            if retry_count != 0:
+                time.sleep(RETRY_DELAY_IN_SEC)
             data = self._get_async_request_data(request_id)
             if data['STATUS'] == 'SUCCESS':
                 return data['response']['identity_key']
             elif data['STATUS'] == 'PROCESSING':
+                retry_count += 1
                 continue
             else:
                 raise AuthError(data['message'])
 
+        if retry_count == MAX_RETRIES:
+            raise RuntimeError(f"Max retry limit reached to get third party identity created.")
+
     @handleError
     def get_third_party_identity_key_by_name(self, integration_key, name):
         url = "/integrations/{0}/identities".format(integration_key)
@@ -785,7 +824,8 @@ class MammothConnector(object):
         return ds_id
 
     def wait_till_ds_ready(self, ds_id):
-        while True:
+        retry_count = 0
+        while retry_count < MAX_RETRIES*4:
             datasets = self.list_datasets()
             ds_ready = False
             for i in range(len(datasets)):
@@ -793,15 +833,18 @@ class MammothConnector(object):
                 log.debug(ds)
                 if ds['id'] == ds_id:
                     if ds['status'] == 'processing':
-                        time.sleep(2)
+                        time.sleep(RETRY_DELAY_IN_SEC)
                         continue
                     if ds['status'] in ['ready', 'error']:
-                        time.sleep(1)
+                        time.sleep(RETRY_DELAY_IN_SEC)
                         ds_ready = True
                         break
                     log.debug(ds['status'])
             if ds_ready:
                 break
+            retry_count += 1
+        if retry_count == MAX_RETRIES*4:
+            raise RuntimeError(f"Max time limit reached to get ds ready.")
 
     @handleError
     def apply_template_to_view(self, view_id, template_config, wait_till_ready=False):
@@ -810,39 +853,42 @@ class MammothConnector(object):
         data = r.json()
 
         if wait_till_ready:
-            time.sleep(5)
             self.wait_till_view_is_ready(view_id)
         return data['STATUS'] == 'PROCESSING'
 
     @handleError
     def track_future_id(self, future_id):
         """
-        Future request tracker that keeps making call to get resouces(corelist)
-        untill chnages are detected against the future id
+        Future request tracker that send GET request to Future API
+        and returns response as JSON object when status is not in 
+        processing (i.e. SUCCESS/FAILED)
         """
-        make_api_call = True
-        response = None
-        while make_api_call:
-            time.sleep(1)
-            resource_resp = self._make_signed_request('get', '/resources')
-            resource_response_json = resource_resp.json()
-            if resource_response_json['future_request']['changed'] == True:
-                for item in resource_response_json['future_request']['items']:
-                    if item['id'] == future_id:
-                        log.debug("The response item against future id {} is {}".format(future_id, item))
-                        response = item['response']
-                        make_api_call = False
-        log.debug("The future response is {}".format(response))
-        return response
+        if not isinstance(future_id, int):
+            raise RuntimeError(f"Future id should be a numeric value")
 
+        response = None
+        retry_count = 0
+        while retry_count < MAX_RETRIES:
+            if retry_count != 0:
+                time.sleep(RETRY_DELAY_IN_SEC)
+            resource_resp = self._make_signed_request('get', f'/future/{future_id}')
+            resource_resp_json = resource_resp.json()  
+            future_object = resource_resp_json['future']
+            if future_object['status'] != 'processing':
+                response = future_object['response']
+                break
+            retry_count += 1
+
+        if retry_count == MAX_RETRIES:
+            raise RuntimeError(f"Max limit reached to future request response.")
+        return response
 
     @handleError
     def add_explore_card(self, view_id, params):
         """
         Method to add explore card
-        Make post request toderivatives api, fetch derivative id,
-        Make post request to queries api, fetch query id,
-        Make get request to queries api, fetch data against the derivative
+        Make post request to derivatives api, fetch derivative id,
+        Make get request to derivative api, fetch data of the derivative
         :param view_id:
         :type view_id:
         :param params:
@@ -857,21 +903,63 @@ class MammothConnector(object):
             return add_derivative_response
         derivative_id = add_derivative_response['id']
         query_response = {'STATUS': 'PROCESSING'}
-        while query_response['STATUS'] != 'SUCCESS':
-            query_params = {'CONDITION': None}
-            response = self._make_signed_request('post', '/workspaces/{0}/derivatives/{1}/queries'.format(view_id, derivative_id),
-                                                    json=query_params)
-            query_response = response.json()
-            log.debug("Query response is {}".format(query_response))
-            time.sleep(2)
-
-        query_id = query_response['query_id']
-        data_response = {'STATUS': 'PROCESSING'}
-        while data_response['STATUS'] != 'READY':
-            response = self._make_signed_request('get', '/workspaces/{0}/derivatives/{1}/queries/{2}'.format(view_id, derivative_id,
-                                                                                                                query_id))
+        retry_count = 0
+        data_response = query_response
+        while data_response['STATUS'] != 'READY' and retry_count < MAX_RETRIES:
+            response = self._make_signed_request('post', f'/workspaces/{view_id}/derivatives/{derivative_id}', json={})
             data_response = response.json()
             log.debug("Data response is {}".format(data_response))
-            time.sleep(2)
+            if retry_count != 0:
+                time.sleep(RETRY_DELAY_IN_SEC)
+            retry_count += 1
+        if retry_count == MAX_RETRIES:
+            raise RuntimeError(f"Max limit reached to read query data.")
+
+        return data_response
+
+    @handleError
+    def add_batch_columns(self, ds_id, param):
+        """
+        Method to add batch columns to the Dataset
+        :param ds_id:
+        :param param - List of batch column/s to be added to the dataset
+        """
+        response = self._make_signed_request('patch', f'/datasets/{ds_id}',
+                                                json={"patch": [{"op": "replace", "path": "batch_columns","value": {"add_columns": param,"remove_columns": []}}]})
+        return response.json()
+
+    @handleError
+    def get_batch_column_keys(self, ds_id):
+        """
+        Method to get batch columns of the dataset
+        :param ds_id:
+        :return - List of batch column/s keys
+        """
+        list_batch_columns = []
+        metadata = self.get_dataset(ds_id)['meta_data']
+        for value in metadata:
+            if value['internal_name'] in RESERVED_BATCH_COLUMN_INTERNAL_NAMES_AND_KEYS.keys():
+                batch_col_key = RESERVED_BATCH_COLUMN_INTERNAL_NAMES_AND_KEYS[value['internal_name']]
+                batch_columns_list.append(batch_col_key)
+
+        return list_batch_columns
+
+    @handleError
+    def remove_batch_columns(self, ds_id, rem_cols):
+        """
+        Method to remove batch columns from the Dataset
+        :param ds_id:
+        :param rem_cols - List of batch column/s (keys) to be removed from the dataset
+        """
+        # Get batch columns of the dataset
+        batch_columns_list = self.get_batch_column_keys(ds_id)
+        add_cols = list(set(batch_columns_list) - set(rem_cols))
+
+        response = self._make_signed_request('patch', '/datasets/{0}'.format(ds_id),
+                                             json={"patch": [{"op": "replace", "path": "batch_columns",
+                                                              "value": {"add_columns": add_cols,
+                                                                        "remove_columns": rem_cols}}]})
+        return response.json()
+
+
 
-        return data_response
\ No newline at end of file
